import numpy as np
from typing import (
    Optional,
    TYPE_CHECKING
)

from .enum import Color

if TYPE_CHECKING:
    from .board import Board


class Territory:
    """Represents a territory i.e. a list of nearby vertices of the same color

    :Parameters:
        x: Optional[:class:`int`]
            The x coordinate of the vertice to use to initiate the territory. This cannot be mixed with the `vertices` parameter
        y: Optional[:class:`int`]
            The y coordinate of the vertice to use to initiate the territory. This cannot be mixed with the `vertices` parameter
        vertices: Optional[list[tuple[:class:`int`, :class:`int`]]]
            A list of vertices to initiate the territory. This cannot be mixed with the `x` and `y` parameters

    :Notes:
        When using the `x, y` initializer, the territory is built by exploring the board while with `vertices` it only uses the given vertices without any exploration

    :Raises:
        :class:`TypeError`
            Parameters are not of the right type
        :class:`ValueError`
            Failed to create the territory with the given parameters"""
    def __init__(self, *,
                 x: Optional[int] = None,
                 y: Optional[int] = None,
                 vertices: Optional[list[tuple[int, int]]] = None,
                 board: 'Board'
                 ):
        self._board: Board = board
        self._vertices: list[tuple[int, int]]
        self._color: Color

        if vertices is not None and x is None and y is None:
            if not isinstance(vertices, list):
                raise TypeError(f'Expected type list for vertices but got {vertices.__class__.__name__}')
            if not vertices:
                raise ValueError('Vertices list cannot be empty')
            if any(board[v] is Color.Empty for v in vertices):
                raise ValueError('Cannot create a territory with empty vertices')
            self._color = board[vertices[0]]
            if any(board[v] is not self._color for v in vertices):
                raise ValueError('Vertices are of different colors')
            self._vertices = vertices
            if not self.is_coherent:
                raise ValueError('Vertices are not all nearby')
        elif x is not None and y is not None and vertices is None:
            if board[x, y] is Color.Empty:
                raise ValueError('Attempting to create a Territory from an empty vertice')
            self._vertices = self._explore(x, y)
            self._color = board[x, y]
        else:
            raise TypeError("Please provide either vertices or both x and y")

    def __repr__(self):
        return f"<{self.__class__.__name__} board={self._board} size={self.size} color={self._color}>"

    @property
    def size(self) -> int:
        """Returns the number of vertices in the territory"""
        return len(self._vertices)

    def clone(self, board: Optional['Board'] = None) -> 'Territory':
        """Returns a deep copy of the territory

        :Parameters:
            board: Optional[:class:`Board`]
                The board to link the new territory, if None it's the same as the current territory

        :Returns:
            The copy of the territory
        """
        new_territory = Territory(vertices=list(self._vertices), board=board if board else self._board)
        return new_territory

    def is_coherent(self) -> bool:
        return set(self._explore(*self._vertices[0])) == set(self._vertices)

    def _explore(self, x: int, y: int) -> list[tuple[int, int]]:
        to_explore = set(self._board.around(x, y))
        explored = [(x, y)]
        while to_explore:
            for i, j in list(to_explore):
                if self._board[i, j] is self._board[x, y]:
                    explored.append((i, j))
                    to_explore.update({k for k in self._board.around(i, j) if k not in explored})
                to_explore.remove((i, j))
        return explored

    @classmethod
    def merge(cls,
              *territories: 'Territory',
              with_vertice: Optional[tuple[int, int]] = None
              ) -> 'Territory':
        """Merge several connected territories into one

        :Parameters:
            \\*territories: :class:`Territory`
                An argument list of territories to merge
            with_vertice: Optional[tuple[:class:`int`, :class:`int`]]
                A vertice to connect all territories, if none is specified, territories have to be already connected

        :Raises:
            :class:`TypeError`
                Parameters aren't of the right type
            :class:`ValueError`
                Failed to merge territories

        :Returns:
            The new territory"""
        if n := next((t for t in territories if not isinstance(t, Territory)), None):
            raise TypeError(f'Expected value of type Territory, but got {n.__class__.__name__}')
        if len(territories) < 2:
            raise TypeError('Expected at least 2 argument for merging')
        if any(t._board != territories[0]._board for t in territories):
            raise ValueError('Territories are not on the same board')
        vertices = []
        if with_vertice:
            if not isinstance(with_vertice, (tuple, list, np.ndarray)) or not len(with_vertice) == 2:
                raise TypeError('Expected 2-len tuple for with_vertices')
            vertices.append(with_vertice)
        for x in territories:
            vertices.extend(x._vertices)
        new_territory = cls(vertices=list(set(vertices)), board=territories[0]._board)
        if not new_territory.is_coherent:
            raise ValueError('Territories are not nearby')
        return new_territory

    def is_nearby(self, t: 'Territory') -> bool:
        """Checks if a territory is connected

        :Parameters:
            t: :class:`Territory`
                The territory to check

        :Returns:
            Indicate if the territory is connected or not"""
        return any(t.is_touching(x, y) for x, y in self._vertices)

    def is_touching(self, x: int, y: int) -> bool:
        """Checks if a vertice is touching the territory

        :Parameters:
            x: :class:`int`
                The x coordinate of the vertice
            y: :class:`int`
                The y coordinate of the vertice

        :Returns:
            Indicate if the vertice is touching the territory"""
        for i, j in self._board.around(x, y):
            if (i, j) in self._vertices:
                return True
        return False

    def includes(self, x: int, y: int, color: Optional[Color] = None) -> bool:
        """Checks if a vertice is included in the territory

        :Parameters:
            x: :class:`int`
                The x coordinate of the vertice
            y: :class:`int`
                The y coordinate of the vertice
            color: :class:`Color`
                The color of the targeted vertice

        :Returns:
            Indicates if the vertice is included or not"""
        return (color is None or color is self._color) and (x, y) in self._vertices

    def _update(self, x: int, y: int, color: Color) -> None:
        if self.is_touching(x, y):
            if color is self._color and (x, y) not in self._vertices:
                self._vertices.append((x, y))

    @property
    def board(self) -> 'Board':
        """The board associated with the territory"""
        return self._board

    @property
    def color(self) -> Color:
        """The color of the territory"""
        return self._color

    def freedom(self) -> list[tuple[int, int]]:
        """Calculate the freedom of the territory, i.e. the vertices where it can expend

        :Returns:
            The list of available vertices to expend the territory"""
        return self._hypothetical_freedom()

    def _hypothetical_freedom(self,
                              x: Optional[int] = None,
                              y: Optional[int] = None,
                              color: Optional[Color] = None
                              ) -> list[tuple[int, int]]:
        free_vertices = []
        if any(k is None for k in [x, y, color]) and not all(k is None for k in [x, y, color]):
            raise TypeError('x, y and color have to be all specified')
        if x is not None and y is not None and color is not None:
            hypothetical_board = np.copy(self._board._grid)
            hypothetical_board[x, y] = color
            if color == self._color and self.is_touching(x, y):
                hypothetical_vertices = self._vertices+[(x, y)]
            else:
                hypothetical_vertices = self._vertices
        else:
            hypothetical_board = self._board._grid
            hypothetical_vertices = self._vertices
        for x, y in hypothetical_vertices:
            for i, j in self._board.around(x, y):
                if hypothetical_board[i, j] is Color.Empty and (i, j) not in free_vertices:
                    free_vertices.append((i, j))
        return free_vertices



__title__ = 'gogame'
__author__ = 'BlackRaven0405'
__license__ = 'MIT'
__copyright__ = 'Copyright 2022-present BlackRaven0405'
__version__ = '0.0.1'


from .board import *
from .territory import *
from .player import *
from .enum import *



from setuptools import setup
import re

version = ''
with open('gogame/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('version is not set')

readme = ''
with open('README.rst') as f:
    readme = f.read()

setup(
    name='gogame',
    version=version,
    license='MIT',
    description='An easy way to simulate and automize go-like programs.',
    long_description=readme,
    long_description_content_type="text/x-rst",
    author="BlackRaven0405",
    author_email='ljouhault@crowbots.shop',
    packages=['gogame'],
    url='https://github.com/BlackRaven0405/gogame',
    keywords='go gogame',
    python_requires='>=3',
    install_requires=[
          'numpy',
      ],

)


from abc import ABC, abstractmethod
import functools
from typing import (
    Optional,
    TYPE_CHECKING
)

import numpy as np

from .enum import Color

if TYPE_CHECKING:
    from .territory import Territory
    from .board import Board


def in_game(func):
    @functools.wraps(func)
    def f(self, *args, **kwargs):
        if self._in_game:
            if self._color and self._board:
                return func(self, *args, **kwargs)
            else:
                raise Exception("Player is in game but is not fully initialized")
        else:
            raise Exception("Player is not currently in game")
    return f


class Player(ABC):
    """Represents a go player.
    This class has to be overridden to implement the :func:`play()` method

    :Parameters:
        name: Optional[:class:`str`]
            The name of the player (only used to identify it)
    """
    def __init__(self, name: Optional[str] = None):
        self._in_game: bool = False
        self._board: Optional[Board] = None
        self._color: Optional[Color] = None
        self.name = name

    def __str__(self):
        if self.name:
            return self.name
        else:
            return self.__repr__()

    def __repr__(self):
        return f"<{self.__class__.__name__} color={self._color} board={self._board}>"

    def _initiate(self, board: 'Board', color: Color):
        self._in_game = True
        self._board = board
        self._color = color

    def _clear_state(self):
        self._in_game = False
        self._board = None
        self._color = None

    @property
    def is_in_game(self) -> bool:
        """Check if the player is currently linked to a board"""
        return self._in_game

    @property
    def board(self) -> Optional['Board']:
        """Returns the associated board if any"""
        return self._board

    @in_game
    def get_board_matrix(self) -> np.ndarray:
        """Returns the current state of the board (0 for empty, 1 for the player, -1 for the opponent)"""
        return self._board.matrix(self._color)

    @in_game
    def playable_moves(self) -> list[tuple[int, int]]:
        """Returns the list of all vertices available for playing"""
        return self._board.playable_moves(self._color)

    @in_game
    def my_vertices(self) -> list[tuple[int, int]]:
        """Returns a list of all vertices owned by the player"""
        return self._board.vertices(self._color)

    @in_game
    def opponent_vertices(self) -> list[tuple[int, int]]:
        """Returns a list of all vertices owned by the opponent"""
        return self._board.vertices(Color.Black if self._color is Color.White else Color.White)

    @in_game
    def free_vertices(self) -> list[tuple[int, int]]:
        """Returns a list of all empty vertices on the board"""
        return self._board.vertices(Color.Empty)

    @in_game
    def my_territories(self) -> list['Territory']:
        """Returns a list of territories owned by the player"""
        return self._board.territories(self._color)

    @in_game
    def opponent_territories(self) -> list['Territory']:
        """Returns a list of territories owned by the opponent"""
        return self._board.territories(Color.Black if self._color is Color.White else Color.White)

    @in_game
    def territories(self) -> list['Territory']:
        """Returns all territories on the board"""
        return self._board.territories()

    @abstractmethod
    def play(self) -> Optional[tuple[int, int]]:
        """This method has to be overridden by subclasses
        It's called when the player has to play and must return a 2-tuple (x, y) representing a move, or None for skipping the turn"""
        pass



import numpy as np
import warnings
import time
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap
from typing import (
    Optional,
    Union,
    Generator,
    TYPE_CHECKING
)

from .territory import Territory
from .enum import Color

if TYPE_CHECKING:
    from .player import Player

cmap = ListedColormap(["white", (0.59, 0.44, 0.2), "black"])


class Board:
    """ Represents the goban of a game

    :Parameters:
        size: Union[:class:`int`, tuple[:class:`int`, :class:`int`]]
            The size of the board, either an int for a square board, or a tuple (height, width)
        show: :class:`int`
            Indicates if the board should be displayed after each move. Default to `False`

    :Notes:
        Board vertices can be accessed through indices:

        >>> b = Board()
        >>> b[0,0]
        <Color.Empty: 0>
    """
    def __init__(self, *, size: Union[int, tuple[int, int]] = 19, show: bool = False):
        if isinstance(size, int):
            height, width = size, size
        elif isinstance(size, (tuple, list, np.ndarray)):
            if len(size) == 2:
                height, width = size
            else:
                raise TypeError(f"size must be a 2-tuple but is of len {len(size)}")
        else:
            raise TypeError(f"size must be of type int or tuple but is of type {size.__class__.__name__}")
        self.show: bool = show
        self._grid: np.ndarray = np.full((height, width), Color.Empty)
        self._last_grid: np.ndarray = np.copy(self._grid)
        self._next_player: Color = Color.Black
        self._prisoners: dict[Color, int] = {Color.Black: 0, Color.White: 0}
        self._territories: list[Territory] = []
        self._players: dict[Color, Optional[Player]] = {Color.Black: None, Color.White: None}

    def __getitem__(self, name: tuple[int, int]) -> Color:
        if not (isinstance(name, tuple) and len(name) == 2):
            raise IndexError("Not a valid indice")
        return self._grid[name]

    def __repr__(self):
        return f"<{self.__class__.__name__} width={self._grid.shape[0]} height={self._grid.shape[1]}>"

    @staticmethod
    def other_player(color: Color) -> Color:
        """Returns the other color (black for white and white for black)"""
        if color is Color.Empty:
            raise ValueError("Player cannot be Empty")
        return Color.Black if color is Color.White else Color.White

    def join(self, player: 'Player') -> None:
        """Links a player to the board for a game

        :Parameters:
            player: :class:`Player`
                The player to link

        :Raises:
            :class:`ValueError`
                There are already two players linked"""
        if not self._players[Color.Black]:
            self._players[Color.Black] = player
            player._initiate(self, Color.Black)
        elif not self._players[Color.White]:
            self._players[Color.White] = player
            player._initiate(self, Color.White)
        else:
            raise ValueError("Both player are already joined")

    def remove_player(self, player: 'Player') -> None:
        """Unlinks a player from the board

        :Parameters:
            player: :class:`Player`
                The player to unlink

        :Raises:
            :class:`ValueError`
                This player is not linked to the board"""
        if player is self._players[Color.Black]:
            self._players[Color.Black]._clear_state()
            self._players[Color.Black] = None
        elif player is self._players[Color.White]:
            self._players[Color.White] = None
            self._players[Color.White]._clear_state()
        else:
            raise ValueError("Player is not joined")

    def clear_players(self) -> None:
        """Unlinks all players from the board"""
        for player in self._players.values():
            player._clear_state()
        self._players = {Color.Black: None, Color.White: None}

    @classmethod
    def from_grid(cls, grid: np.ndarray):
        """Initialize a board from an 2D array of :class:`Color`

        :Parameters:
            grid: :class:`np.ndarray`
                A 2D array of :class:`Color` representing the state of the board

        :Returns:
            The new created board"""
        new_board = cls()
        new_board._grid = grid
        new_board._last_grid = np.full(grid.shape, Color.Empty)
        new_board._init_territories()
        return new_board

    def clone(self) -> 'Board':
        """Returns a deep copy of the board"""
        new_board = Board()
        new_board.show = self.show
        new_board._grid = np.copy(self._grid)
        new_board._next_player = self._next_player
        new_board._prisoners = dict(self._prisoners)
        new_board._territories = [t.clone(new_board) for t in self._territories]
        return new_board

    def _init_territories(self) -> None:
        self._territories = []
        for x in range(self._grid.shape[0]):
            for y in range(self._grid.shape[1]):
                if self[x, y] is not Color.Empty and not any((x, y) in t._vertices for t in self._territories):
                    self._territories.append(Territory(x=x, y=y, board=self))

    def display(self) -> None:
        """Displays the board as a numpy matrix"""
        plt.imshow(self.matrix(Color.Black), cmap=cmap, vmin=-1, vmax=1)
        plt.pause(0.1)

    def is_playable(self, x: int, y: int, color: Color) -> bool:
        """
        Checks if a move is valid

        :Parameters:
            x: :class:`int`
                The x coordinate to check
            y: :class:`int`
                The y coordinate to check
            color: :class:`int`
                The color of the player to check

        :Returns:
            Indicates if the move is valid
        """
        if self[x, y] is not Color.Empty:
            return False
        grid = np.copy(self._grid)
        grid[x, y] = color
        if np.all(grid is self._last_grid):
            if np.all(self._last_grid is self._grid):
                return True
            return False
        mine = []
        opponent = []
        for i, j in self.around(x, y):
            if t := self.get_territory(i, j):
                if t.color is color:
                    mine.append(t)
                else:
                    opponent.append(t)
        if any(not t._hypothetical_freedom(x, y, color) for t in opponent):
            return True
        if len(opponent) == len(list(self.around(x, y))):
            return False
        if mine and all(not t._hypothetical_freedom(x, y, color) for t in mine):
            return False
        return True

    def playable_moves(self, color: Color) -> list[tuple[int, int]]:
        """ Gives the list of valid move for a given color

        :Parameters:
            color: :class:`Color`
                The player

        :Returns:
            A list of all vertices where the player can play
        """
        playable = []
        for x in range(self._grid.shape[0]):
            for y in range(self._grid.shape[1]):
                if self.is_playable(x, y, color):
                    playable.append((x, y))
        return playable

    def run_game(self, max_turn: Optional[int] = 1000, max_duration: Optional[int] = None) -> 'Player':
        """Runs a game on this board between two players. The players have to be linked to the board with :func:`join` before

        :Parameters:
            max_turn: Optional[:class:`int`]
                The maximum number of move before ending the game
            max_duration: Optional[:class:`int`]
                The maximum number of seconds before ending the game

        :Raises:
            :class:`ValueError`
                Not enough players to start the game
            :class:`TypeError`
                A player returns an invalid move type

        :Returns:
            The player who wins the game"""
        if any(player is None for player in self._players):
            raise ValueError("The board needs two players to be run")
        if max_turn is None and max_duration is None:
            warnings.warn("max_turn and max_duration are both to None, game might run forever")
        c = 0
        starting_time = time.time()
        while (not c or c < max_turn) and (not max_duration or time.time()-starting_time < max_duration):
            move = self._players[self._next_player].play()
            if move is None:
                if self.skip(player=self._next_player):
                    return self.winner()
            elif isinstance(move, (tuple, list, np.ndarray)) and len(move) == 2:
                self.play(*move, color=self._next_player)
            else:
                raise TypeError("play method must return None or a 2-tuple")
            c += 1
        return self.winner()

    def play(self, x: int, y: int, *, color: Color) -> None:
        """Play a move manually without using Player object

        :Parameters:
            x: :class:`int`
                The x coordinate of the move to play
            y: :class:`int`
                The y coordinate of the move to play
            color: :class:`Color`
                The color of the move to play

        :Raises:
            :class:`ValueError`
                The move is invalid, or it's the wrong player"""
        if color is not self._next_player:
            raise ValueError('Wrong player')

        if not self.is_playable(x, y, color):
            raise ValueError('You cannot play here')

        self._last_grid = np.copy(self._grid)
        self._grid[x, y] = color
        self._next_player = self.other_player(self._next_player)

        modified = [t for t in self.territories(color) if t.is_touching(x, y)]
        if len(modified) == 1:
            modified[0]._update(x, y, color)
        elif len(modified) >= 2:
            merge_territory = Territory.merge(*modified, with_vertice=(x, y))
            for t in modified:
                self._territories.remove(t)
            self._territories.append(merge_territory)

        for t in self.territories(self.other_player(color)):
            if not t.freedom:
                self._territories.remove(t)
                for i, j in t._vertices:
                    self._grid[i, j] = 0
                    self._prisoners[color] += 1

        if not any(t.includes(x, y, color) for t in self._territories):
            new_territory = Territory(x=x, y=y, board=self)
            self._territories.append(new_territory)
        if self.show:
            self.display()

    def skip(self, *, player: Color, show: Optional[bool] = None) -> Optional[bool]:
        """Skip a turn manually without using Player object

            :Raises:
                :class:`ValueError`
                    It's the wrong player"""
        if player is not self._next_player:
            raise ValueError('Wrong player')
        if np.all(self._last_grid is self._grid):
            return True
        self._next_player = self.other_player(player)
        if show is True or (show is None and self.show):
            self.display()

    def winner(self) -> 'Player':
        """Returns the current winner of the board by comparing the scores of both player
        In case of equality, White wins

        :Returns:
            The player who currently leads the game"""
        if self.score(Color.Black) > self.score(Color.White):
            return self._players[Color.Black]
        return self._players[Color.White]

    def around(self,
               x: int,
               y: int,
               include_center: bool = False
               ) -> Generator[tuple[int, int], None, None]:
        """A quick way to get vertices around a given point

        :Parameters:
            x: :class:`int`
                The x coordinate of the point
            y: :class:`int`
                The y coordinate of the point
            include_center: :class:`bool`
                Wether to include the given point or not

        :Yields:
            The points around"""
        if x > 0:
            yield x - 1, y
        if y > 0:
            yield x, y - 1
        if x < self._grid.shape[0] - 1:
            yield x + 1, y
        if y < self._grid.shape[1] - 1:
            yield x, y + 1
        if include_center:
            yield x, y

    def prisoners(self, color: Color) -> int:
        """Get the number of prisoners owned by a player

        :Parameters:
            color: :class:`Color`
                The color of the player

        :Returns:
            The number of prisoners"""
        return self._prisoners[color]

    def matrix(self, color: Color) -> np.ndarray:
        """Returns the current state of the board as a numpy matrix

        * 0 is an empty vertice
        * 1 is a vertice the player owns
        * -1 is a vertice owned by the opponent

        :Parameters:
            color: :class:`Color`
                The color of the player

        :Returns:
            The matrix representing the board
        """
        grid = np.vectorize(lambda x: x.value)(self._grid)
        return -grid if color is Color.White else grid

    def territories(self, color: Optional[Color] = None) -> list[Territory]:
        """Returns territories currently on the board. If a color is specified, only territories of the given color are returned

        :Parameters:
            color: Optional[:class:`Color`]
                The color of the territories to get

        :Returns:
            A list of territories"""
        if color is None:
            return self._territories
        else:
            return [x for x in self._territories if x.color is color]

    def get_territory(self,
                      x: int,
                      y: int
                      ) -> Optional[Territory]:
        """Get a territory from a vertice

        :Parameters:
            x: :class:`int`
                The x coordinate of the territory to get
            y: :class:`int`
                The y coordinate of the territory to get

        :Returns:
            The territory which owns the vertice if any"""
        for t in self._territories:
            if t.includes(x, y):
                return t
        return None

    def vertices(self, color: Color) -> list[tuple[int, int]]:
        """Get all vertices from a given color

        :Parameters:
            color: :class:`Color`
                The color of the vertices to get

        :Returns:
            The list of vertices"""
        vertices = []
        for x in range(self._grid.shape[0]):
            for y in range(self._grid.shape[1]):
                if self[x, y] is color:
                    vertices.append((x, y))
        return vertices

    def score(self, color: Color) -> int:
        """Returns the score of a player i.e. the number of vertices belonging to the player + the number of his prisoners

        :Parameters:
            color: :class:`Color`
                The color of the player

        :Returns:
            The score of the given player"""
        return self._prisoners[color] + np.count_nonzero(self._grid is color)
